#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

import os
import shutil
import warnings
import tempfile
import numpy as np


# TIRL IMPORTS

from tirl import settings as ts


class Buffer(object):
    """
    Buffer class. Uniform interface for memory-mapped and in-memory arrays.
    Automatically deletes linked files on the hard disk when discarded. This
    class is a successor of the Buffer namedtuple that was used in the original
    implementation of Domain.

    """
    def __init__(self, arr=None, fname=None, file_no=None, signature=None):

        # Signature
        self.signature = signature

        # File ID and file name
        if isinstance(file_no, int):
            self.file_no = file_no
        else:
            self.file_no = None
        if isinstance(fname, str):
            self.fname = fname
        else:
            self.fname = None

        # Infer file name from memory map
        if isinstance(arr, np.memmap) and (fname is None):
            self.fname = arr.filename

        # Data (even if None)
        self.data = arr

    def __del__(self):
        if (self.file_no is not None) and os.path.isfile(self.fname):
            try:
                os.close(self.file_no)
            except Exception:
                pass
            try:
                os.remove(self.fname)
            except Exception:
                warnings.warn(
                    "Temporary file could not be removed from {}."
                        .format(self.fname))

    def __bool__(self):
        """ The implicit bool value of the Buffer object is False, when no data
        is attached to it. """
        return self.data is not None

    def __array__(self):
        """ Array interface of the Buffer class. """
        return self.data

    def copy(self):
        if self.data is None:
            data = None
        elif isinstance(self.data, np.memmap):
            old_file = self.data.filename
            fd, new_file = tempfile.mkstemp(dir=ts.TWD, prefix="buffer_")
            with open(old_file, mode="rb+") as of:
                with open(new_file, mode="rb+") as nf:
                    shutil.copyfileobj(of, nf)
            data = np.memmap(new_file, mode="r+", dtype=self.data.dtype,
                             offset=0, shape=self.data.shape, order="C")
        else:
            data = self.data.copy()
        return Buffer(data, self.fname, self.file_no)

    @classmethod
    def _load(cls, dump):
        file_no = dump.get("file_no")
        fname = dump.get("fname")
        data = dump.get("data")
        ret = Buffer(arr=data, fname=fname, file_no=file_no)
        return ret

    def _dump(self):
        dbuffer = {
            "type": ".".join([self.__class__.__module__,
                              self.__class__.__name__]),
            "id": str(id(self)),
            "file_no": self.file_no,
            "fname": self.fname,
            "data": self.data,
        }
        return dbuffer
