#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

import joblib
import inspect
import numpy as np
import collections


# TIRL IMPORTS

from tirl import exceptions as te

# DEFINITIONS

CACHE_MODES = {"memory", "db"}


# IMPLEMENTATION

class Cache(object):
    """ Cache object for instance-specific use. Usage: define Cache object
    as part of the initialisation, then call the three public methods (query,
    retrieve, store) in succession from the same function/method.

    Example:
        signature = ("John", "Doe", "42", "Sheffield")  # for hashing
        key, exists = self.cache.query(signature)
        if exists:
            value = self.cache.retrieve(key)
        else:
            value = "successful person"
            self.cache.store(key, value)

    Note:
        To access the same cache database from multiple functions/methods,
        specify the same caller ID (the name of the function/method) while
        using query(), retrieve(), and store().

    """

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ INITIALISATION ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

    def __init__(self, mode="memory", maxsize=2):
        """
        Initialisation of Cache object.

        :param mode: Storage mode. "memory": in RAM, "db": database.
        :type mode: Union["memory", "db"]

        """
        # Validate input for Cache Mode
        if not mode.lower() in CACHE_MODES:
            raise te.ConstructionError(
                    "Unrecognised GImage mode: '{0}'. Available Cache modes "
                    "are: {1}".format(mode, ", ".join(CACHE_MODES)))
        else:
            self.mode = mode

        # Validate input for maxsize
        if isinstance(maxsize, (int, np.integer)):
            if maxsize > 0:
                self.maxsize = maxsize
            else:
                raise te.ConstructionError(
                    "Maximum cache size must be a positive integer.")
        else:
            raise te.ConstructionError(
                "Maximum cache size must be a positive integer.")

        # Create cache and accession frequency database
        if mode == "memory":
            self.cache = dict()
            self.freqs = dict()
        else:
            # TODO
            raise NotImplementedError()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PUBLIC METHODS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

    def query(self, *hashables, caller=None):
        """
        Queries Cache object for existing entry based on hashable
        inputs.

        """
        # Calculate hash for hashables
        key = joblib.hash(hashables)

        # Identify caller function/method
        if caller is None:
            # caller = str(inspect.stack()[1][0].f_code.co_name)
            caller = "default"

        try:
            fcache = self.cache[caller]
            freq = self.freqs[caller]
        except:
            if self.mode == "memory":
                fcache = self.cache[caller] = dict()
                freq = self.freqs[caller] = collections.defaultdict()
            else:
                # TODO
                raise NotImplementedError()

        # Return: key, exists
        return key, key in fcache.keys()

    def retrieve(self, key, caller=None):
        """
        Retrieve cached data based on key.

        :param key: Access key (return value of Cache.query).
        :type key: str

        :returns: Cached data.

        """
        # Identify caller function/method
        if caller is None:
            caller = str(inspect.stack()[1][0].f_code.co_name)
        fcache = self.cache[caller]
        freq = self.freqs[caller]

        # Retrieve data from cache
        try:
            result = fcache[key]
        except KeyError:
            raise te.CacheError("Retrieve was unsuccessful for '{0}' "
                                "using key '{1}'.".format(caller, key))
        else:
            # Increase accession count for the specific key
            try:
                freq[key] += 1
            except:
                freq[key] = 1

        return result

    def store(self, key, value, caller=None):
        """ Adds an expensive result to the cache under a particular key. """

        # Identify caller function/method
        if caller is None:
            caller = str(inspect.stack()[1][0].f_code.co_name)
        fcache = self.cache[caller]
        freq = self.freqs[caller]

        # Add data to cache
        try:
            # Delete from cache when full
            if len(fcache) >= self.maxsize:
                delkey = min(freq.items(), key=lambda f: f[1])[0]
                fcache.pop(delkey)
                freq.pop(delkey)

            # Store new value
            fcache[key] = value

            # Increase accession count for the new key
            try:
                freq[key] += 1
            except:
                freq[key] = 1

        except KeyError:
            raise te.CacheError("The requested data could not be stored in "
                                "the instance's cache.")


if __name__ == "__main__":
    print("""This module is not intended for execution.""")

