#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

import numpy as np


# TIRL IMPORTS

from tirl import exceptions as te, settings as ts
from tirl.operations.tensor import TensorOperator


# IMPLEMENTATION

class TensorBroadcast(TensorOperator):
    """
    TensorBroadcast (child of TensorOperator) -
        broadcasting a tensor field to match a specified tensor shape.

    """
    def __init__(self, target_shape, n_cpu=ts.CPU_CORES, name=None):
        """
        Initialisation of TensorBroadcast.

        :param target_shape: Target tensor shape.
        :type target_shape: tuple[int]

        """
        super(TensorOperator, self).__init__(
            self.__operation, chunker=None, n_cpu=n_cpu, name=name,
            target_shape=target_shape)

    @staticmethod
    def __operation(x, target_shape):
        """
        Broadcast operation for voxel-major oriented chunks.

        :param x: (n_voxels, *tensor_shape) voxel-major oriented chunk
        :type x: np.ndarray
        :param target_shape: target tensor shape
        :type target_shape: tuple[int]

        :return: broadcasted chunk
        :rtype: np.ndarray

        """
        shape = x.shape[1:]
        extra_dims = len(target_shape) - len(shape)
        slicer = (slice(None), *((np.newaxis,) * extra_dims), Ellipsis)
        x = x[slicer]
        if extra_dims:
            for axis, reps in enumerate(target_shape[:extra_dims]):
                x = np.repeat(x, reps, axis=axis + 1)
        else:
            if shape[0] == 1:
                x = np.repeat(x, target_shape[0], axis=1)
            else:
                raise te.BroadcastError("Tensor shape {} is not broadcastable "
                                        "to {}.".format(shape, target_shape))
        return x


if __name__ == "__main__":
    print("This module is not intended for execution.")
