#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

import numpy as np
from operator import mul
from functools import reduce


# TIRL IMPORTS

from tirl import exceptions as te, settings as ts
from tirl.operations.spatial import SpatialOperator


# IMPLEMENTATION

class Robust(SpatialOperator):
    """
    Clips tensor values below and above the element-specific robust range. The
    robust range is defined by percentiles (default: 1 and 99).

    """
    def __init__(self, low=1, high=99, n_cpu=ts.CPU_CORES, name=None,
                 **opkwargs):
        """
        Initialisation of Robust.

        :param low:
        :param high:
        :param n_cpu:
        :param name:
        :param opkwargs:
        """

        # Verify input
        if isinstance(low, (int, np.integer)):
            if (low < 0) or (low > 100):
                raise ValueError("Lower percentile must be in range [0,100].")
        else:
            raise TypeError("Lower percentile must be an integer value.")
        if isinstance(high, (int, np.integer)):
            if (high < 0) or (high > 100):
                raise ValueError("Higher percentile must be in range [0,100].")
        else:
            raise TypeError("Higher percentile must be an integer value.")
        if low >= high:
            raise te.ArgumentError("Lower percentile must not be greater than "
                                   "or equal to the higher percentile.")
        self.low = low
        self.high = high
        operation = self._clip
        opargs = ()
        super(Robust, self).__init__(operation, *opargs, radius=1, n_cpu=n_cpu,
                                     name=name, **opkwargs)

    def _clip(self, chunk, *unused_args, **opkwargs):
        n_channels = reduce(mul, self.tshape)
        a_min = opkwargs.get("a_min")
        a_min = (a_min,) * n_channels if not hasattr(a_min, "__iter__") \
            else a_min
        a_max = opkwargs.get("a_max")
        a_max = (a_max,) * n_channels if not hasattr(a_max, "__iter__") \
            else a_max
        result = np.zeros_like(chunk)
        if self.taxes:
            for i in range(n_channels):
                indices = np.unravel_index(i, self.tshape)
                slicer = tuple(slice(int(ix), int(ix) + 1) for ix in indices)
                result[slicer] = \
                    np.clip(chunk[slicer], a_min=a_min[i], a_max=a_max[i])
        else:
            result = np.clip(chunk, a_min=a_min[0], a_max=a_max[0])
        return result

    def __call__(self, tfield, out=None, **opkwargs):
        # Calculate clipping values from percentiles
        low = np.percentile(tfield.data, q=self.low, axis=tfield.vaxes)
        high = np.percentile(tfield.data, q=self.high, axis=tfield.vaxes)
        # Set clipping values
        a_min = opkwargs.pop("a_min", None) or low.ravel()
        a_max = opkwargs.pop("a_max", None) or high.ravel()
        opkwargs.update({"a_min": a_min, "a_max": a_max})
        self.opkwargs.update(opkwargs)
        self.tdim = tfield.tdim
        self.tshape = tfield.tshape
        self.taxes = tfield.taxes
        # Perform clipping
        return super(Robust, self).__call__(tfield, out=out, **self.opkwargs)
