#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

import numpy as np
import scipy.sparse as sp
from functools import reduce


# TIRL IMPORTS

from tirl.cache import Cache
from tirl.regularisers.txregulariser import TransformationRegulariser


# IMPLEMENTATION

class DiffusionRegulariser(TransformationRegulariser):
    """
    DiffusionRegulariser - penalises the gradient magnitude of a dense
    displacement field.

    """
    RESERVED_KWARGS = ("transformation", "weight", "logger", "metaparameters")

    def __init__(self, transformation, weight=1.0, logger=None,
                 **metaparameters):
        """
        Initialisation of DiffusionRegularisation.

        :param transformation:
            Transformation object whose parameters are used to compute to the
            regularisation term.
        :type transformation: Transformation or OptimisationGroup
        :param weight: Regularisation weight.
        :type weight: Real
        :param logger:
            If None (default), logs will be redirected to the RootLogger
            instance. You may alternatively specify a different Logger instance
            or its name.
        :type logger: None or str or Logger
        :param kwargs:
            Additional keyword arguments required to compute the
            regularisation term.
        :type kwargs: Any

        """
        if not hasattr(transformation, "domain") or \
            not transformation.domain.iscompact:
            raise TypeError(f"Expected non-linear transformation on a compact "
                            f"domain, got {transformation} instead.")

        # Call parent-class initialisation
        super(DiffusionRegulariser, self).__init__(
            transformation, weight=weight, logger=logger, **metaparameters)

        self.cache = Cache(maxsize=1)

    def function(self):
        """
        Calculates the total scalar regularisation cost.

        """
        field = self.transformation.std().absolute().field()
        grad = np.stack(np.gradient(field.data, axis=field.vaxes), axis=-1)
        del field
        return np.sum(grad ** 2)

    def grad1(self):
        """
        Calculates the total derivative of the regularisation cost with respect
        to the parameters. Note that this is a Gateaux-derivative, that is
        co-dependent on the field update v: since R = U'D'DU,
        dR[u;v] = <$\Delta$u, v>.

        """
        vectors = self.transformation.vectors(fullstd=True)
        L = sp.block_diag([self.grad2()] * vectors.shape[-1])
        grad = L @ vectors.T.reshape(-1, 1)
        return grad.ravel()

    def grad2(self):
        """
        Calculates the total second derivative of the regularisation cost with
        respect to the parameters.

        Note that this quantity is independent from the current field or its
        update: since R = U'D'DU, the second derivative is the Laplace
        operator, which is returned in discrete (grid) form, as a sparse array
        of shape (l*m*n, l*m*n) or (l*m, l*m). The letters l, m, and n denote
        the shape of the transformation domain.

        """
        shape = self.transformation.domain.shape
        signature = (shape,)  # for hashing
        key, exists = self.cache.query(signature, caller="grad2")
        if exists:
            laplace = self.cache.retrieve(key, caller="grad2")
        else:
            laplace = discrete_laplace_operator(*shape, order="C")
            # Apply correction for Neumann boundary condition
            diag = np.asarray(np.sum(laplace, axis=0)).ravel() + 2 * len(shape)
            laplace = sp.dia_matrix(laplace)
            laplace.setdiag(-diag)
            laplace = self.weight * laplace
            self.cache.store(key, laplace, caller="grad2")

        return laplace


def discrete_laplace_operator(*dims, order="C"):
    """
    Creates a sparse discrete Laplace operator for a (dim0, dim1, ...) grid.

    :param dims: grid dimensions
    :type dims: int

    :returns: sparse discrete Laplace operator

    """

    # Check input
    dims = (dims,) if not isinstance(dims, (tuple, list)) else tuple(dims)
    if not all(isinstance(dim, (int, np.integer)) for dim in dims):
        raise ValueError("Dimensions of the N-dimensional domain must be "
                         "integers.")
    ndim = len(dims)
    size = np.product(dims)
    if not ndim >= 1:
        raise ValueError("Laplace operator must be at least 1-dimensional.")

    # Revert dimensions for Fortran-order
    if order.upper() == "F":
        dims = dims[::-1]

    terms = []
    for current_index, dim in enumerate(dims):
        dim = int(dim)
        ones = np.ones(dim)
        D = sp.spdiags([ones, -2*ones, ones], [-1, 0, 1], dim, dim)
        if current_index > 0:
            pre = reduce(lambda a, b: sp.kron(a, b),
                         map(lambda c: sp.identity(dims[c]),
                             range(0, current_index)))
            L = sp.kron(pre, D)
        else:
            L = D
        if current_index < ndim - 1:
            for i in range(current_index + 1, ndim):
                L = sp.kron(L, sp.identity(int(dims[i])))
        assert L.shape == (size, size)
        terms.append(L)
    else:
        result = reduce((lambda a, b: a + b), terms)

    return result
