#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

import numpy as np
from numbers import Real


# TIRL IMPORTS

from tirl.regularisers.txregulariser import TransformationRegulariser
from tirl.transformations.transformation import Transformation


# IMPLEMENTATION

class NormTxRegulariser(TransformationRegulariser):
    """
    NormTxRegulariser - calculates regularisation term from the norm of
    transformation parameters.

    """
    RESERVED_KWARGS = ("transformation", "order", "weight", "logger",
                       "metaparameters")

    def __init__(self, transformation, order=2, weight=1.0, logger=None,
                 **metaparameters):
        """
        Initialisation of NormTxRegulariser.

        :param transformation:
            Transformation object whose parameters are used to compute to the
            regularisation term.
        :type transformation: Transformation or OptimisationGroup
        :param weight: Regularisation weight.
        :type weight: Real
        :param logger:
            If None (default), logs will be redirected to the RootLogger
            instance. You may alternatively specify a different Logger instance
            or its name.
        :type logger: None or str or Logger
        :param kwargs:
            Additional keyword arguments required to compute the
            regularisation term.
        :type kwargs: Any

        """
        # Call superclass initialisation
        super(NormTxRegulariser, self).__init__(
            transformation, weight=weight, logger=logger, **metaparameters)

        # Set class-specific metaparameters
        self.order = order

    @property
    def order(self):
        return self.metaparameters.get("order")

    @order.setter
    def order(self, r):
        if isinstance(r, Real):
            self.metaparameters.update(order=float(r))
        else:
            raise TypeError(f"Expected real number for norm oder, "
                            f"got {r.__class__.__name__}")

    def function(self):
        """
        Returns p-norm of the transformation parameters.

        """
        p = self.order
        res = np.sum(np.abs(self.transformation.parameters[:]) ** p) ** (1. / p)
        res /= self.transformation.parameters.size
        return res

    def grad1(self):
        """
        Analytical Jacobian of the p-norm.

        """
        p = self.order
        x = self.parameters
        pnorm = self.function()
        jac = x * np.abs(x) ** (p - 2) / (pnorm ** (p - 1))
        return jac

    def grad2(self):
        """
        Analytical Hessian of the p-norm.

        """
        p = self.order
        n = self.parameters.size
        if p == 1:
            # Avoid calculations and rounding errors: Hessian is 0 for L1-norm
            hess = np.zeros((n, n))
        else:
            pnorm = self.function()
            x = self.parameters.reshape((-1, 1))
            xx = x * np.abs(x) ** (p - 2)
            hess = (1 - p) * pnorm ** (1 - 2 * p) * xx @ xx.T \
                   + (p - 1) * np.diag(x) ** (p - 2) / (pnorm ** (p - 1))
        return hess


if __name__ == "__main__":
    print("""This module is not intended for execution.""")
