#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar
# Date: 10 June 2020


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


from tirl.scripts.mnd import __version__
__tirlscript__ = True


# DEPENDENCIES

import os
import sys
import argparse


# TIRL IMPORTS

import tirl
import tirl.settings as ts


def combine(args):
    """ Main program code. """

    histo = tirl.load(args.histo)
    block = tirl.load(args.block)
    slice_ = tirl.load(args.slice)
    mri = tirl.load(args.vol)
    histo.domain.chain += block.domain.chain[2:] + slice_.domain.chain[1:]
    result = mri.evaluate(histo.domain)

    # Save output
    outputdir = os.path.dirname(args.out)
    if not os.path.isdir(outputdir):
        os.makedirs(outputdir)
    ext = ts.EXTENSIONS["TImage"]
    result.save(args.out + f".{ext}")
    ext = ts.TIMAGE_DEFAULT_SNAPSHOT_EXT
    result.snapshot(args.out + f".{ext}")


def create_cli():
    """
    Creates command-line interface.

    """
    prog = "combine"
    usage = "combine --histo <timg> --block <timg> --slice <timg> --vol <timg>"
    descr = "Combines the transformations from stages 1-3 of the MND " \
            "registration pipeline to achieve an end-to-end mapping between " \
            "histology and MRI coordinates."
    parser = argparse.ArgumentParser(prog, usage, description=descr)
    parser.add_argument("--histo", metavar="timg", type=str, required=True,
                        default="", help="Registered histology TImage.")
    parser.add_argument("--block", metavar="timg", type=str, required=True,
                        default="", help="Registered tissue block TImage.")
    parser.add_argument("--slice", metavar="timg", type=str, required=True,
                        default="", help="Registered brain slice TImage.")
    parser.add_argument("--vol", metavar="timg", type=str, required=True,
                        default="", help="MRI volume as TImage.")
    parser.add_argument("--out", metavar="basename", type=str, required=True,
                        default="", help="Basename for output.")

    return parser


def main(*args):
    """ Main program code. """

    parser = create_cli()
    if args:
        args = parser.parse_args(args)
        combine(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main(*sys.argv[1:])
