#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _       __      __  _         _
#  |__   __| |_   _| |  __ \  | |      \ \    / / (_)       (_)
#     | |      | |   | |__) | | |       \ \  / /   _   ___   _    ___    _ __
#     | |      | |   |  _  /  | |        \ \/ /   | | / __| | |  / _ \  | '_ \
#     | |     _| |_  | | \ \  | |____     \  /    | | \__ \ | | | (_) | | | | |
#     |_|    |_____| |_|  \_\ |______|     \/     |_| |___/ |_|  \___/  |_| |_|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

import numpy as np


# TIRL IMPORTS

from tirl import settings as ts
from tirl.tirlvision import tfield as tvtf
from skimage.exposure import rescale_intensity

# Safe import Matplotlib
if ts.ENABLE_VISUALISATION:
    import os
    import matplotlib
    if os.getenv("DISPLAY"):
        matplotlib.use(ts.MPL_BACKEND)
    else:
        matplotlib.use("agg")
    import matplotlib.pyplot as plt
else:
    import warnings
    warnings.warn("Visualisations are disabled in tirlconfig. "
                  "Matplotlib was not loaded.", ImportWarning)


# IMPLEMENTATION

def create_object_figure(obj):
    try:
        return obj.figure, True
    except AttributeError:
        obj.figure = plt.figure()
        return obj.figure, False


def show_image(obj, **kwargs):
    interrupt = kwargs.pop("interrupt", True)
    fig, ready = create_object_figure(obj)
    c = obj.costs[0]
    # target = c.filter(c.target)
    # source = c.filter(c.source)
    target = c.target
    source = c.source
    source = source.evaluate(target.domain)
    diff = source - target
    imdata = rescale_intensity(diff.tensors[0].data, out_range=(0, 1))
    if ready:
        plt.ion()
        obj.ax.set_data(imdata)
        plt.ioff()
        plt.pause(0.01)
    else:
        ax = fig.add_subplot(111)
        obj.ax = ax.imshow(imdata, cmap="gray")
        plt.pause(0.01)
        fig.show()


def show_deformation(obj, **kwargs):
    fig, ready = create_object_figure(obj)
    if ready:
        # field = obj.transformation.deformation()
        # shape = obj.transformation.domain.shape
        # field = TField.fromarray(
        #     field.reshape(*shape, -1), tensor_axes=-1, order="V")
        field = obj.transformation.field()
        tvtf.preview_hsv(field, frame=obj.ax, **kwargs)
    else:
        shape = obj.transformation.domain.shape
        ax = fig.add_subplot(111)
        obj.ax = ax.imshow(np.zeros(shape))
        plt.pause(0.01)
        fig.show()
